@@ -68,13 +68,17 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
   if (var.estim == "sandwich" & method == "dwols") {stop("Sandwich variance estimation only available with G-estimation.\n")}
   if (class(treat.mod) == "formula") {treat.mod <- list(treat.mod); blip.mod <- list(blip.mod); tf.mod <- list(tf.mod)}
   if (length(treat.mod) != length(blip.mod) | length(treat.mod) != length(tf.mod)) {stop("-treat.mod-, -blip.mod- and -tf.mod- must all have same length.")}
-  if (truncate <0 | truncate > 0.5) {stop("-truncate- must lie between 0 and 0.5.")}
+  if (truncate < 0 | truncate > 0.5) {stop("-truncate- must lie between 0 and 0.5.")}
   # infer stages from treat.mod
   K <- length(treat.mod)
   obj$K <- K
   # if no data set, create one from the models
   if (is.null(data)) {
-    data <- cbind(outcome,do.call("cbind",lapply(blip.mod,get_all_vars)),do.call("cbind",lapply(treat.mod,get_all_vars)),do.call("cbind",lapply(tf.mod,get_all_vars)))
+    # form list of variables
+    model.list <- c(blip.mod, tf.mod, treat.mod)
+    # extract variables: note we drop any intercept-only terms
+    data <- cbind(outcome, do.call('cbind', lapply(model.list[which(model.list != '~1')], get_all_vars)))
+    # remove duplicates
     data <- data[!duplicated(lapply(data,summary))]
     obj$outcome <- outcome
   } else {
@@ -144,9 +148,9 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
 
 	if (missing == 'ipcw') {
 	keep <- c(1:N)
-      keep <- keep[!(keep %in% new.drop)]
+      keep <- keep[!(keep %in% new.drop)]				
 	} else {
-      keep <- keep[!(keep %in% drop)]
+      keep <- keep[!(keep %in% drop)]		
 	}
 
       # only work with those we want to kep at this stage
@@ -477,7 +481,7 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
 
 # printing
 #' @export
-print.DTRreg <- print.DWSurv <- function(x,...) {
+print.DTRreg <- function(x,...) {
 	K <- x$K
 	# longest blip variable name lengths, if applicable, get length of blip excluding treatment terms
 	max.length <- 0
@@ -563,7 +567,7 @@ print.DTRreg <- print.DWSurv <- function(x,...) {
 	}
 }
 #' @export
-summary.DTRreg <- summary.DWSurv <- function(object,...) {
+summary.DTRreg <- function(object,...) {
 	print.DTRreg(object)
 }
 
@@ -601,7 +605,7 @@ predict.DTRreg <- function(object, newdata, treat.range=NULL, ...) {
 	return(obj)
 }
 #' @export
-coef.DTRreg <- coef.DWSurv  <- function(object,...) {
+coef.DTRreg <- function(object,...) {
 	x <- object
 	K <- x$K
 	obj <- x$psi
@@ -613,7 +617,7 @@ coef.DTRreg <- coef.DWSurv  <- function(object,...) {
 }
 # confidence interval
 #' @export
-confint.DTRreg  <- confint.DWSurv <- function(object, parm = NULL, level = 0.95, type = "se", ...) {
+confint.DTRreg <- function(object, parm = NULL, level = 0.95, type = "se", ...) {
   x <- object
   K <- x$K
   obj <- x$psi
@@ -651,10 +655,10 @@ confint.DTRreg  <- confint.DWSurv <- function(object, parm = NULL, level = 0.95,
 
 # diagnostics
 #' @export
-plot.DTRreg <- plot.DWSurv <- function(x, ...) {
+plot.DTRreg <- function(x, method = "DTRreg", ...) {
   # original Y is in x$obs.Y
   # fitted Y (one for each stage) is in x$fitted
-  if(class(x) == "DTRreg"){
+  if(method == "DTRreg"){
     cat("Y residuals (for treatment-free and blip models):\n")
     for (j in 1:x$K) {
       advance <- readline("Hit <Return> to see next plot:")
@@ -677,7 +681,7 @@ plot.DTRreg <- plot.DWSurv <- function(x, ...) {
       cat("Stage",j,"\n")
       plot(x$treat.mod.fitted[[j]],sub.caption=paste("Stage",j,"treatment"))
     }
-  }else{
+  }else if(method == "DWSurv"){
     cat("Y residuals (for treatment-free and blip models):\n")
     for (j in 1:x$K) {
       advance <- readline("Hit <Return> to see next plot:")
@@ -838,8 +842,6 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
   K <- length(blip.mod)
   obj$K <- K
   obj$optimization <- optimization
-  # defined counters if bootstrap normal or empirical
-  if(var.estim == "bootstrap" & boot.opt != "standard") obj$rej.boot <- vector(mode = "list", length = K)
   # checking input validity (continued)
   for(j in 1:K){
     if(class(treat.mod[[j]]) != "formula") stop(paste("The stage ", j, " treatment model must be supplied as a formula with the treatment variable on the left hand side.", sep = ""))
@@ -847,7 +849,11 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
     if(class(time[[j]]) != "formula" | length(all.vars(time)) > 1) stop("The stage ", j, " survival time must be supplied as a formula with nothing on the left hand side and the survival time variable on the right hand side.")
   }
   #time <- list(time); treat.mod <- list(treat.mod); blip.mod <- list(blip.mod); tf.mod <- list(tf.mod); cens.mod <- list(cens.mod)
-
+  
+  # verify if cens.mod supplied as no censoring
+  nocens <- FALSE
+  if(length(cens.mod[[1]])==2) nocens <- TRUE
+  
   # if no data set, create one from the models
   if (is.null(data)) {
     data <- cbind(do.call("cbind",lapply(time,get_all_vars)), do.call("cbind",lapply(blip.mod,get_all_vars)),do.call("cbind",lapply(treat.mod,get_all_vars)),do.call("cbind",lapply(tf.mod,get_all_vars)), do.call("cbind",lapply(cens.mod,get_all_vars)))
@@ -859,7 +865,7 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
     for(j in 1:K){
       obj$time[[j]] <- data[,which(colnames(data) == all.vars(time[[j]]))]
     }
-    obj$status <- data[,which(colnames(data) == all.vars(cens.mod[[1]])[1])]
+    obj$status <- data[,which(colnames(data) == all.vars(cens.mod[[1]])[1])] 
   }
   # make sure data is a data frame (can't be a matrix)
   data <- data.frame(data)
@@ -867,7 +873,7 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
   obj$data <- data
   Y <- obj$time
   delta <- obj$status
-
+  
   if (any(is.na(delta))) { # missing survival times will be normal if patients don't reach stage
     stop("Missing values in status are not allowed.")
   }
@@ -887,7 +893,7 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
     Hpsi <- model.matrix(blip.mod[[j]], model.frame(blip.mod[[j]], dataetaj, na.action = 'na.pass'))
     Hbeta <- model.matrix(tf.mod[[j]], model.frame(tf.mod[[j]], dataetaj, na.action = 'na.pass'))
     Halpha <- model.matrix(treat.mod[[j]], model.frame(treat.mod[[j]], dataetaj, na.action = 'na.pass'))
-    if(is.null(cens.mod[[j]]) != TRUE) {
+    if(nocens == FALSE) {
       Hlambda <- model.matrix(cens.mod[[j]], model.frame(cens.mod[[j]], dataetaj, na.action = 'na.pass'))
       drop <- append(drop,dataetaj$id[which(apply(is.na(cbind(A, Hpsi, Hbeta, Halpha, Hlambda)), 1, any))])
     }else{
@@ -899,7 +905,7 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
       Y[[j]][which(data$id %in% drop)] <- 0
     }
   }
-
+  
   # keep track of id removed
   obj$drop <- drop
   # keep track of who enter stages
@@ -916,27 +922,27 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
     Yj <- Ycumul[eta[[j]] == 1]
     dataetaj <- data[eta[[j]] == 1, ]
     deltaj <- delta.store[[j]] <- delta[eta[[j]] == 1]
-
+    
     # get data: Hpsi = blip, Hbeta = treatment-free, Halpha = treatment, Hlambda = censoring
     A <- A.store[[j]] <- as.numeric(model.response(model.frame(treat.mod[[j]], dataetaj, na.action = 'na.pass')))
     Hpsi <- model.matrix(blip.mod[[j]], model.frame(blip.mod[[j]], dataetaj, na.action = 'na.pass'))
     Hbeta <- model.matrix(tf.mod[[j]], model.frame(tf.mod[[j]], dataetaj, na.action = 'na.pass'))
     Halpha <- model.matrix(treat.mod[[j]], model.frame(treat.mod[[j]], dataetaj, na.action = 'na.pass'))
-    if(is.null(cens.mod[[j]]) != TRUE) Hlambda <- model.matrix(cens.mod[[j]], model.frame(cens.mod[[j]], dataetaj, na.action = 'na.pass'))
-
+    if(nocens == FALSE) Hlambda <- model.matrix(cens.mod[[j]], model.frame(cens.mod[[j]], dataetaj, na.action = 'na.pass'))
+    
     # list of censoring and treatment variables
     treat.var <- all.vars(treat.mod[[j]])[1]
-
+    
     # store blip variable names for output
     obj$blip.list[[j]] <- colnames(Hpsi)
-
+    
     # force matrices in case intercept-only models
     Hpsi <- Hpsi.store[[j]] <-  as.matrix(Hpsi)
     Hbeta <- as.matrix(Hbeta)
     Halpha <- Halpha.store[[j]] <- as.matrix(Halpha)
-    if(is.null(cens.mod[[j]]) != TRUE) Hlambda <- Hlambda.store[[j]] <- as.matrix(Hlambda)
+    if(nocens == FALSE) Hlambda <- Hlambda.store[[j]] <- as.matrix(Hlambda)
     obj$n[[j]] <- nrow(dataetaj)
-
+    
     # check if treatments are binary, if not, and dWOLS selected, abort
     A.bin <- as.numeric(length(unique(A)) <= 2)
     if (A.bin == 0) {
@@ -952,26 +958,26 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
     }
     # for compatibility with other functions
     obj$cts[[j]] <- "bin"
-
+    
     # treatment model: binary logistic
     alpha <- glm(A ~ -1 + Halpha, family = binomial(link = "logit"))
     obj$treat.mod.fitted[[j]] <- alpha
     Ahat <- Ahat.store[[j]] <- fitted(alpha)
     obj$obs.treat[[j]] <- A
-
+    
     # censoring model: binary logistic
-    if(sum(deltaj) != length(deltaj) & is.null(cens.mod[[j]]) != TRUE){
+    if(sum(deltaj) != length(deltaj) & nocens == FALSE){
       lambda <- glm(deltaj ~ -1 + Hlambda, family = binomial(link = "logit"))
       obj$cens.mod.fitted[[j]] <- lambda
       Dhat <- Dhat.store[[j]] <- fitted(lambda)
       obj$obs.cens[[j]] <- deltaj
-    }else if(sum(deltaj) == length(deltaj) & is.null(cens.mod[[j]]) != TRUE){
+    }else if(sum(deltaj) == length(deltaj) & nocens == FALSE){
       Dhat <- Dhat.store[[j]] <- deltaj
       cat("Warning: no censoring in stage", j,", censoring model will be ignored.\n")
     }else{
       Dhat <- Dhat.store[[j]] <- deltaj
     }
-
+    
     # estimation step
     # weights
     if (is.function(weight) == FALSE) {
@@ -989,8 +995,8 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
     est <- solve(t(cbind(Hbetaj, Aj * Hpsij)) %*% cbind(wj * Hbetaj, wj * Aj * Hpsij)) %*% t(cbind(Hbetaj, Aj * Hpsij)) %*% (wj * logY)
     psi <- est[(dim(Hbeta)[2] + 1):(dim(Hbeta)[2] + dim(Hpsi)[2])]
     beta <- est[1:dim(Hbeta)[2]]
-
-    # if parametric bootstrap requested, done within each stage
+    
+    # if parametric bootstrap requested
     if(var.estim == "bootstrap" & boot.opt != "standard"){
       psi.boot <- matrix(NA, nrow = B, ncol = length(psi))
       residuals <- logY - Hbetaj %*% beta - Aj * Hpsij %*% psi
@@ -998,49 +1004,21 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
       mean.res <- mean(residuals)
       sd.res <- sd(residuals)
       boot.data <- dataetaj
-      obj$rej.boot[[j]] <- 0
       for (k in 1:B) {
+        boot.data[,which(colnames(boot.data) == all.vars(cens.mod[[1]])[1])] <- rbinom(obj$n[[j]], 1, Dhat.store[[j]])
         if(boot.opt == "empirical"){
-          # resample censoring indicator from glm model
-          boot.data[,which(colnames(boot.data) == all.vars(cens.mod[[1]])[1])] <- rbinom(obj$n[[j]], 1, Dhat.store[[j]])
-          # for rows with delta=1, resample from empirical distribution of residuals
           bins <- with(reshist, sample(length(mids), sum(boot.data[,which(colnames(boot.data) == all.vars(cens.mod[[1]])[1])]), prob = density, replace = TRUE))
           newres <- runif(obj$n[[j]], reshist$breaks[bins], reshist$breaks[bins + 1])
-          boot.data[, which(colnames(boot.data) == all.vars(time[[j]]))] <- exp(Hbeta %*% beta + A*Hpsi %*% psi + newres)
-          # try to fit model
-          mod <- try(DWSurv(time = list(obj$time.mod[[j]]), blip.mod = list(obj$blip.mod[[j]]), treat.mod = list(obj$treat.mod[[j]]), tf.mod = list(obj$tf.mod[[j]]), cens.mod = list(obj$cens.mod[[j]]), boot.data, var.estim="none", quiet = TRUE), silent = TRUE)
-          while(class(mod) != "DWSurv"){
-            # record the number of rejected samples in stage j
-            obj$rej.boot[[j]] <- obj$rej.boot[[j]] + 1
-            boot.data[,which(colnames(boot.data) == all.vars(cens.mod[[1]])[1])] <- rbinom(obj$n[[j]], 1, Dhat.store[[j]])
-            bins <- with(reshist, sample(length(mids), sum(boot.data[,which(colnames(boot.data) == all.vars(cens.mod[[1]])[1])]), prob = density, replace = TRUE))
-            newres <- runif(obj$n[[j]], reshist$breaks[bins], reshist$breaks[bins + 1])
-            boot.data[, which(colnames(boot.data) == all.vars(time[[j]]))] <- exp(Hbeta %*% beta + A*Hpsi %*% psi + newres)
-            mod <- try(DWSurv(time = list(obj$time.mod[[j]]), blip.mod = list(obj$blip.mod[[j]]), treat.mod = list(obj$treat.mod[[j]]), tf.mod = list(obj$tf.mod[[j]]), cens.mod = list(obj$cens.mod[[j]]), boot.data, var.estim="none", quiet = TRUE), silent = TRUE)
-          }
-          psi.boot[k,] <- mod$psi[[1]]
         }else if(boot.opt == "normal"){
-          # resample censoring indicator from glm model
-          boot.data[,which(colnames(boot.data) == all.vars(cens.mod[[1]])[1])] <- rbinom(obj$n[[j]], 1, Dhat.store[[j]])
-          # for rows with delta=1, resample residuals from normal distribution
           newres <- rnorm(obj$n[[j]], mean = mean.res, sd = sd.res)
-          boot.data[, which(colnames(boot.data) == all.vars(time[[j]]))] <- exp(Hbeta %*% beta + A*Hpsi %*% psi + newres)
-          # try to fit model
-          mod <- try(DWSurv(time = list(obj$time.mod[[j]]), blip.mod = list(obj$blip.mod[[j]]), treat.mod = list(obj$treat.mod[[j]]), tf.mod = list(obj$tf.mod[[j]]), cens.mod = list(obj$cens.mod[[j]]), boot.data, var.estim="none", quiet = TRUE), silent = TRUE)
-          while(class(mod) != "DWSurv"){
-            obj$rej.boot[[j]] <- obj$rej.boot[[j]] + 1
-            boot.data[,which(colnames(boot.data) == all.vars(cens.mod[[1]])[1])] <- rbinom(obj$n[[j]], 1, Dhat.store[[j]])
-            newres <- rnorm(obj$n[[j]], mean = mean.res, sd = sd.res)
-            boot.data[, which(colnames(boot.data) == all.vars(time[[j]]))] <- exp(Hbeta %*% beta + A*Hpsi %*% psi + newres)
-            mod <- try(DWSurv(time = list(obj$time.mod[[j]]), blip.mod = list(obj$blip.mod[[j]]), treat.mod = list(obj$treat.mod[[j]]), tf.mod = list(obj$tf.mod[[j]]), cens.mod = list(obj$cens.mod[[j]]), boot.data, var.estim="none", quiet = TRUE), silent = TRUE)
-          }
-          psi.boot[k,] <- mod$psi[[1]]
         }
+        boot.data[, which(colnames(boot.data) == all.vars(time[[j]]))] <- exp(Hbeta %*% beta + A*Hpsi %*% psi + newres)
+        psi.boot[k,] <- DWSurv(time = list(obj$time.mod[[j]]), blip.mod = list(obj$blip.mod[[j]]), treat.mod = list(obj$treat.mod[[j]]), tf.mod = list(obj$tf.mod[[j]]), cens.mod = list(obj$cens.mod[[j]]), boot.data, var.estim="none", quiet = TRUE)$psi[[1]]
       }
       obj$covmat[[j]] <- var(psi.boot)
       obj$psi.boot[[j]] <- psi.boot
     }
-
+    
     # use estimates to identify optimal treatments
     # depending on optimization goal
     if (optimization == "max") {
@@ -1062,7 +1040,7 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
     Yj <- exp(log(Yj) + obj$log.regret[[j]])
     old.Ycumul <- Ycumul
     Ycumul[eta[[j]] == 1] <- Yj
-
+    
     # Asymptotic variance estimator done stage-by-stage
     if (var.estim == "asymptotic" & is.function(weight) == FALSE) {
       logY <- as.matrix(log(old.Ycumul[eta[[j]] == 1]))
@@ -1086,7 +1064,7 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
       dU.dalpha <- (1/nj) * (crossprod(deltaj * abs(A - Ahat)/(deltaj * Dhat + (1 - deltaj) * (1 - Dhat)) * (1 - 2*A) * Ahat * (1 - Ahat) * residuals * Htheta, Halpha))
       U.adj <- U.adj.store[[j]] <- U.adj - t(dU.dalpha %*% solve(ds.dalpha) %*% t(score.alpha))
       # terms that depends on lambda censoring model
-      if(is.null(cens.mod[[j]]) != TRUE){
+      if(nocens == FALSE){
         Hlambda <- as.matrix(Hlambda)
         score.lambda <- score.lambda.store[[j]] <- (deltaj - Dhat) * Hlambda
         ds.dlambda <- ds.dlambda.store[[j]] <- (-1/nj) * (crossprod(Dhat * (1 - Dhat) * Hlambda, Hlambda))
@@ -1116,7 +1094,7 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
           # treatment component
           dUjp1.dpsijp1 <- dUjp1.dpsijp1 - t(Htheta.store[[l]]) %*% Halpha.store[[l]] %*% solve(ds.dalpha.store[[l]]) %*% t(score.alpha.store[[l]]) %*% (-1/njp1 * delta.store[[l]] * w.store[[l]] * (1 - 2 * A.store[[l]]) * Ahat.store[[l]] * (1 - Ahat.store[[l]]) * Hpsitemp)
           # censoring component if necessary
-          if(is.null(cens.mod[[j]]) != TRUE){
+          if(nocens == FALSE){
             dUjp1.dpsijp1 <- dUjp1.dpsijp1 - t(Htheta.store[[l]]) %*% Hlambda.store[[l]] %*% solve(ds.dlambda.store[[l]]) %*% t(score.lambda.store[[l]]) %*% (1/njp1 * delta.store[[l]] * abs(A.store[[l]] - Ahat.store[[l]])/(delta.store[[l]] * Dhat.store[[l]] + (1 - delta.store[[l]]) * (1 - Dhat.store[[l]]))^2 * (2 * delta.store[[l]] - 1) * Dhat.store[[l]] * (1 - Dhat.store[[l]]) * Hpsitemp)
           }
           dUjp1.dpsijp1 <- dUjp1.dpsijp1[(length(obj$beta[[l]]) + 1):(length(obj$beta[[l]]) + length(obj$psi[[l]])),]
@@ -1130,7 +1108,7 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
       dUadj.dtheta <- dU.dtheta <- (-1/nj) * (crossprod(HD, Htheta))
       if(asymp.opt == "adjusted"){
         dUadj.dtheta <- dUadj.dtheta  - t(Htheta) %*% Halpha %*% solve(ds.dalpha) %*% t(score.alpha) %*% (-1/nj * deltaj * w * (1 - 2 * A) * Ahat * (1 - Ahat) * Htheta)
-        if(is.null(cens.mod[[j]]) != TRUE){
+        if(nocens == FALSE){
           dUadj.dtheta <- dUadj.dtheta - t(Htheta) %*% Hlambda %*% solve(ds.dlambda) %*% t(score.lambda) %*% (1/nj * deltaj * abs(A - Ahat)/(deltaj * Dhat + (1 - deltaj) * (1 - Dhat))^2 * (2 * deltaj - 1) * Dhat * (1 - Dhat) * Htheta)
         }
         invdUadj.dtheta <- solve(dUadj.dtheta)
@@ -1147,44 +1125,25 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
       }
     }
   }
-
+  
   # optimal outcome relevant only for observed events
   obj$Y.opt <- Ycumul
-
+  
   # standard errors
   # standard bootstrap
   if (var.estim == "bootstrap" & boot.opt == "standard") {
     psi.boot <- list()
     M <- nrow(data)
-    # keep track of rejected samples
-    obj$rej.boot <- 0
     for (i in 1:B) {
-      # resample rows of data with replacement
       data.boot <- data[sample(1:M, replace=TRUE),]
-      mod <- try(DWSurv(time = obj$time.mod, blip.mod = obj$blip.mod, treat.mod = obj$treat.mod, tf.mod = obj$tf.mod, cens.mod = obj$cens.mod, data.boot, var.estim="none", quiet = TRUE), silent = TRUE)
-      # resample as long as the model fails with the bootstrap sample
-      while(class(mod) != "DWSurv"){
-        obj$rej.boot <- obj$rej.boot + 1
-        data.boot <- data[sample(1:M, replace=TRUE),]
-        mod <- try(DWSurv(time = obj$time.mod, blip.mod = obj$blip.mod, treat.mod = obj$treat.mod, tf.mod = obj$tf.mod, cens.mod = obj$cens.mod, data.boot, var.estim="none", quiet = TRUE), silent = TRUE)
-      }
-      psi.boot[[i]] <- mod$psi
+      psi.boot[[i]] <- DWSurv(time = obj$time.mod, blip.mod = obj$blip.mod, treat.mod = obj$treat.mod, tf.mod = obj$tf.mod, cens.mod = obj$cens.mod, data.boot, var.estim="none", quiet = TRUE)$psi
     }
     psi.boot <- do.call(function(...) mapply(rbind, ..., SIMPLIFY=FALSE), psi.boot)
     covmat <- lapply(psi.boot, var)
     obj$covmat <- covmat
     obj$psi.boot <- psi.boot
-    # print number of rejected bootstrap resamples
-    cat("Warning:", obj$rej.boot,"bootstrap resamples were rejected.\n")
   }
-
-  # if bootstrap empirical or normal, print the number of rejected samples by stage
-  if(var.estim == "bootstrap" & boot.opt != "standard"){
-    for(j in 1:K){
-      cat("Warning:", obj$rej.boot[[j]],"bootstrap resamples were rejected in stage", j, ".\n")
-    }
-  }
-
+  
   # if variance requested, calculate non-regularity
   if (var.estim != "none" & A.bin == 1) {
     for (j in K:1) {
@@ -1202,6 +1161,232 @@ DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, wei
     }
   }
   # return
-  class(obj) <- "DWSurv"
+  class(obj) <- "DTRreg"
   obj
 }
+
+###########################################################
+### Generalzied dynamic weighted ordinary least squares ###
+###########################################################
+
+############################################
+### G-dWOLS function for multiple stages ###
+############################################
+
+# outcome: specifies the outcome variable
+# Xpsi1: tailoring variables that interact with A in the blip function
+# Xpsi2: tailoring variables that interact with A^2 in the blip funciton
+# blip function is such that gamma = psi1*Xpsi1*A + psi2*Xpsi2*A^2
+# treat.mod: specifies the treatment model (in terms of covariates)
+# treat.fam: family objects for treatment model (with link function, as specified in the standard glm function)
+#            Note: this argument is only relevant if a weight function based on the continuous doses is used
+#            ex: gaussian(link = "identity")
+#                Gamma(link = "log")
+#                inverse.gaussian(link = "1/mu^2")
+#                poisson(link = "log")
+# tf.mod: specifies the treatment-free model
+# weight.fcn: specifies the weight function to be used
+# data: specifies the data
+# m: specifies the number of groups for weight functions based on binned doses
+# k: specifies the number of stages
+# treat.range: range of permissible treatment values.
+
+
+gdwols<-function(outcome, Xpsi1, Xpsi2, treat.mod, treat.fam=NULL, tf.mod, weight.fcn, data = NULL, m, k, treat.range=NULL){ 
+  
+  obj <- list()
+  
+  # if no data specified, create data frame
+  if (is.null(data)) {
+    # form list of variables
+    model.list <- c(Xpsi1, Xpsi2, tf.mod, treat.mod)
+    # extract variables: note we drop any intercept-only terms
+    data <- cbind(outcome, do.call('cbind', lapply(model.list[which(model.list != '~1')], get_all_vars)))
+    # remove duplicates
+    data <- data[!duplicated(lapply(data,summary))]
+    Y <- outcome
+  } else {
+    # if data is specified, pull outcome from it
+    Y <- eval(substitute(outcome),envir=data)
+  }
+  
+  # if no treat.fam specified, use Gaussian by default
+  if(is.null(treat.fam)){treat.fam<-gaussian(link = "identity")}
+  
+  # if no treat.range specified, use min/max observed treatment (across all stages)
+  if(is.null(treat.range)){
+    treat.k<-NULL
+    for(j in 1:k){treat.k<-c(treat.k,model.response(model.frame(treat.mod[[j]],data)))}
+    treat.range<-c(min(treat.k),max(treat.k))
+    cat("Warning: no treatment range specified, [min,max] of observed treatment used.\n")
+  }
+  
+  # start at final stage and work backwards
+  for (j in k:1) {
+    
+    # treatment 
+    A <- model.response(model.frame(treat.mod[[j]],data))
+    
+    # different options for determining weights
+    w.vec<-NULL
+    if(weight.fcn=="ipw"){w.vec<-ipw.fcn(A,treat.mod[[j]],treat.fam,data,m)}
+    if(weight.fcn=="cipw"){w.vec<-ipw.cap.fcn(A,treat.mod[[j]],treat.fam,data,m)}
+    if(weight.fcn=="qpom"){w.vec<-q.pom.fcn(A,treat.mod[[j]],data,m)}
+    if(weight.fcn=="wo"){w.vec<-new.bin.o.fcn(A,treat.mod[[j]],data,m)}
+    
+    # outcome model (Y) = treatment free model + blip model
+    # tailoring variables in blip function
+    Hpsi1 <- model.matrix(Xpsi1[[j]], model.frame(Xpsi1[[j]], data))
+    Hpsi2 <- model.matrix(Xpsi2[[j]], model.frame(Xpsi2[[j]], data))
+    psi.dim<-ncol(Hpsi1)+ncol(Hpsi2)
+    # Hpsi <- model.matrix(blip.mod[[j]], model.frame(blip.mod[[j]], data))
+    # variables in treatment free model
+    Hbeta <- model.matrix(tf.mod[[j]], model.frame(tf.mod[[j]], data))	
+    beta.dim <- ncol(Hbeta)
+    # outcome model
+    out <- lm(Y~0+Hbeta+A:Hpsi1+I(A^2):Hpsi2,weights=w.vec)
+    # treatment-free parameter estimates 
+    beta <- as.numeric(out$coef[1:beta.dim])
+    # blip parameter estimates 
+    psi <- as.numeric(out$coef[(beta.dim+1):(beta.dim+psi.dim)])
+    
+    # identify optimal treatments
+    # note: for simplificity, we considered a quadratic blip function of the form 
+    # PSI %*% (X1*A, X2*A^2), e.g. (psi1*X1)*A + (psi2*X2)*A^2, 
+    # where X1 and X2 are (subsets) of the available covariates X
+    # the optimal treatment is thus simply a.opt=-(psi1*X1)/2(psi2*X2) (see paper for more details)
+    opt.step<-(-1/2)*(Hpsi1 %*% psi[1:ncol(Hpsi1)])/(Hpsi2 %*% psi[(ncol(Hpsi1)+1):psi.dim])
+    # NOTE: a.opt is only a maximum if (psi2*X2)<0, otherwise endpoint (only if treatment range is finite)
+    if(abs(min(treat.range)) != Inf & abs(max(treat.range)) != Inf){
+      Hpsi.min <- cbind((Hpsi1*min(treat.range)),(Hpsi2*(min(treat.range)^2)))
+      Hpsi.max <- cbind((Hpsi1*max(treat.range)),(Hpsi2*(max(treat.range)^2)))
+      endpoints.diff<-(Hpsi.min %*% psi)-(Hpsi.max %*% psi)
+      opt <- (as.numeric(Hpsi2 %*% psi[(ncol(Hpsi1)+1):psi.dim] < 0)*opt.step)+
+        (as.numeric(Hpsi2 %*% psi[(ncol(Hpsi1)+1):psi.dim] >= 0)*as.numeric(endpoints.diff>=0)*min(treat.range))+
+        (as.numeric(Hpsi2 %*% psi[(ncol(Hpsi1)+1):psi.dim] >= 0)*as.numeric(endpoints.diff<0)*max(treat.range))
+    } else {
+      opt <- opt.step
+      warn.vec<-which(Hpsi2 %*% psi[(ncol(Hpsi1)+1):psi.dim] > 0)
+      if(length(warn.vec)>0 ){
+        cat("Warning: stage",j,"optimal treatment may be a minimum for some observations. \n")
+      }
+      
+    }
+    
+    # optimal dose must fall withing treatment range
+    opt <- pmin(pmax(opt,min(treat.range)),max(treat.range))
+    opt<-as.vector(opt)
+    
+    Hpsi.obs <- cbind(sweep(Hpsi1,1,A,'*'),sweep(Hpsi2,1,(A^2),'*'))
+    Hpsi.opt <- cbind(sweep(Hpsi1,1,opt,'*'),sweep(Hpsi2,1,(opt^2),'*'))
+    # update (pseudo-) Y by adding regret
+    Y <- Y + (Hpsi.opt %*% psi) - (Hpsi.obs %*% psi)
+    
+    # store estimates
+    obj$psi[[j]] <- psi
+    obj$beta[[j]] <- beta
+    obj$opt.treat[[j]] <- opt
+  }
+  
+  return(obj)
+  
+}
+
+
+#########################################################
+### Different weight functions for two stage function ###
+#########################################################
+
+# different weight functions
+
+# IPW 
+ipw.fcn<-function(A,treat.mod,treat.fam,data,m){
+  
+  Xalpha<-model.matrix(treat.mod,model.frame(treat.mod,data))
+  model.output<-glm(A~0+Xalpha,data=data,family=treat.fam)
+  model.disp<-summary(model.output)$dispersion
+  # model.output<-lm(A~0+Xalpha,data=data)
+  
+  if(treat.fam$family=="gaussian"){
+    weight.vec<-1/dnorm(A,mean=model.output$fitted.values,sd=sqrt(model.disp))
+  }
+  
+  if(treat.fam$family=="Gamma"){
+    weight.vec<-1/dgamma(A,shape=1/model.disp,scale=model.output$fitted.values*model.disp)
+  }
+  
+  return(weight.vec)
+  
+}
+
+# capped IPW 
+ipw.cap.fcn<-function(A,treat.mod,treat.fam,data,m){
+  
+  Xalpha<-model.matrix(treat.mod,model.frame(treat.mod,data))
+  model.output<-glm(A~0+Xalpha,data=data,family=treat.fam)
+  model.disp<-summary(model.output)$dispersion
+  # model.output<-lm(A~0+Xalpha,data=data)
+  
+  if(treat.fam$family=="gaussian"){
+    weights<-1/dnorm(A,mean=model.output$fitted.values,sd=sqrt(model.disp))
+  }
+  
+  if(treat.fam$family=="Gamma"){
+    weights<-1/dgamma(A,shape=1/model.disp,scale=model.output$fitted.values*model.disp)
+  }
+  
+  cap<-quantile(weights,0.99)
+  weight.vec<-pmin(weights,cap)
+  
+  return(weight.vec)
+  
+}
+
+
+# quantile binning based on POM
+q.pom.fcn<-function(A,treat.mod,data,m){
+  
+  Xalpha<-model.matrix(treat.mod,model.frame(treat.mod,data))
+  
+  a.binned<-ntile(A, m)
+  a.cat<-as.factor(a.binned)
+  
+  if(ncol(Xalpha)==1){
+    pom<-polr(a.cat~1)
+  }
+  else{
+    pom<-polr(a.cat~Xalpha[,-1])
+  }
+  
+  a.pom.prob<-pom$fitted.values[cbind(seq_along(a.binned), a.binned)]
+  
+  weight.vec<-(1/m)/a.pom.prob
+  
+  return(weight.vec)
+  
+}
+
+
+# overlap weights
+new.bin.o.fcn<-function(A,treat.mod,data,m){
+  
+  Xalpha<-model.matrix(treat.mod,model.frame(treat.mod,data))
+  
+  a.binned<-ntile(A,m)
+  a.cat<-as.factor(a.binned)
+  
+  if(ncol(Xalpha)==1){
+    pom<-polr(a.cat~1)
+  }
+  else{
+    pom<-polr(a.cat~Xalpha[,-1])
+  }
+  
+  a.pom.prob<-pom$fitted.values[cbind(seq_along(a.binned), a.binned)]
+  
+  weight.vec<-(1/a.pom.prob)/apply(1/pom$fitted.values,1,sum)
+  
+  return(weight.vec)
+  
+}
+
